/******************************************************************************
 * Top contributors (to current version):
 *   Aina Niemetz, Andres Noetzli, Mathias Preiner
 *
 * This file is part of the cvc5 project.
 *
 * Copyright (c) 2009-2025 by the authors listed in the file AUTHORS
 * in the top-level source directory and their institutional affiliations.
 * All rights reserved.  See the file COPYING in the top-level source
 * directory for licensing information.
 * ****************************************************************************
 *
 * Template for the Node kind header.
 */

#include "cvc5_public.h"

#ifndef CVC5__KIND_H
#define CVC5__KIND_H

#include <iosfwd>

#include "base/exception.h"
#include "theory/theory_id.h"

namespace cvc5::internal {
namespace kind {

enum class Kind_t
{
  UNDEFINED_KIND = -1, /**< undefined */
  NULL_EXPR,           /**< Null kind */
  // clang-format off
  /* from builtin */
  SORT_TYPE, /**< specifies types of user-declared 'uninterpreted' sorts (1) */
  INSTANTIATED_SORT_TYPE, /**< instantiated uninterpreted sort (2) */
  UNINTERPRETED_SORT_VALUE, /**< the kind of expressions representing uninterpreted sort values; payload is an instance of the cvc5::internal::AbstractValue class (used in models) (3) */
  BUILTIN, /**< the kind of expressions representing built-in operators (4) */
  EQUAL, /**< equality (two parameters only, sorts must match) (5) */
  DISTINCT, /**< disequality (N-ary, sorts must match) (6) */
  VARIABLE, /**< a variable (not permitted in bindings) (7) */
  BOUND_VARIABLE, /**< a bound variable (permitted in bindings and the associated lambda and quantifier bodies only) (8) */
  SKOLEM, /**< a Skolem variable (9) */
  DUMMY_SKOLEM, /**< a Skolem variable (internal only) (10) */
  SEXPR, /**< a symbolic expression (any arity) (11) */
  WITNESS, /**< a witness expression; first parameter is a BOUND_VAR_LIST, second is the witness body (12) */
  TYPE_CONSTANT, /**< a representation for basic types (13) */
  RAW_SYMBOL, /**< a variable that is not quoted in the smt2 printer (internal only) (14) */
  ABSTRACT_TYPE, /**< abstract type (15) */
  APPLY_INDEXED_SYMBOLIC_OP, /**< operator for the bit-vector extract; payload is an instance of the cvc5::internal::GenericOp class (16) */
  APPLY_INDEXED_SYMBOLIC, /**< generic indexed operator, whose arguments are the list of terms corresponding to (symbolic) indices to the operator followed by its normal arguments (17) */
  SORT_TO_TERM, /**< term representing a sort; payload is an instance of the cvc5::internal::SortToTerm class (18) */
  TYPE_OF, /**< the type of an expression, used internally for RARE (19) */

  /* from booleans */
  CONST_BOOLEAN, /**< truth and falsity; payload is a (C++) bool (20) */
  NOT, /**< logical not (21) */
  AND, /**< logical and (N-ary) (22) */
  IMPLIES, /**< logical implication (exactly two parameters) (23) */
  OR, /**< logical or (N-ary) (24) */
  XOR, /**< exclusive or (exactly two parameters) (25) */
  ITE, /**< if-then-else, used for both Boolean and term ITE constructs; first parameter is (Boolean-sorted) condition, second is 'then', third is 'else' and these two parameters must have same base sort (26) */

  /* from uf */
  APPLY_UF, /**< application of an uninterpreted function; first parameter is the function, remaining ones are parameters to that function (27) */
  FUNCTION_TYPE, /**< a function type (28) */
  LAMBDA, /**< a lambda expression; first parameter is a BOUND_VAR_LIST, second is lambda body (29) */
  HO_APPLY, /**< higher-order (partial) function application (30) */
  CARDINALITY_CONSTRAINT_OP, /**< the cardinality constraint operator; payload is an instance of the cvc5::internal::CardinalityConstraint class (31) */
  CARDINALITY_CONSTRAINT, /**< a fixed upper bound on the cardinality of an uninterpreted sort (32) */
  COMBINED_CARDINALITY_CONSTRAINT_OP, /**< the combined cardinality constraint operator; payload is an instance of the cvc5::internal::CombinedCardinalityConstraint class (33) */
  COMBINED_CARDINALITY_CONSTRAINT, /**< a fixed upper bound on the sum of cardinalities of uninterpreted sorts (34) */
  FUNCTION_ARRAY_CONST, /**< the function constant; payload is an instance of the cvc5::internal::FunctionArrayConst class (35) */
  BITVECTOR_UBV_TO_INT, /**< bit-vector conversion to integer; parameter is treated as an unsigned bit-vector (36) */
  BITVECTOR_SBV_TO_INT, /**< bit-vector conversion to integer; parameter is treated as a signed bit-vector (37) */
  INT_TO_BITVECTOR_OP, /**< operator for the integer conversion to bit-vector; payload is an instance of the cvc5::internal::IntToBitVector class (38) */
  INT_TO_BITVECTOR, /**< integer conversion to bit-vector; first parameter is an INT_TO_BITVECTOR_OP, second is an integer term (39) */

  /* from arith */
  ADD, /**< arithmetic addition (N-ary) (40) */
  MULT, /**< arithmetic multiplication (N-ary) (41) */
  NONLINEAR_MULT, /**< synonym for MULT (42) */
  SUB, /**< arithmetic binary subtraction operator (43) */
  NEG, /**< arithmetic unary negation (44) */
  DIVISION, /**< real division, division by 0 undefined (user symbol) (45) */
  DIVISION_TOTAL, /**< real division with interpreted division by 0 (internal symbol) (46) */
  INTS_DIVISION, /**< integer division, division by 0 undefined (user symbol) (47) */
  INTS_DIVISION_TOTAL, /**< integer division with interpreted division by 0 (internal symbol) (48) */
  INTS_MODULUS, /**< integer modulus, division by 0 undefined (user symbol) (49) */
  INTS_MODULUS_TOTAL, /**< integer modulus with interpreted division by 0 (internal symbol) (50) */
  ABS, /**< absolute value (51) */
  DIVISIBLE, /**< divisibility-by-k predicate; first parameter is a DIVISIBLE_OP, second is integer term (52) */
  POW, /**< arithmetic power (53) */
  POW2, /**< arithmetic power of 2 (54) */
  INTS_ISPOW2, /**< test if an integer is a power of 2 (55) */
  INTS_LOG2, /**< integer logarithm base 2 (round down to power of 2) (56) */
  EXPONENTIAL, /**< exponential (57) */
  SINE, /**< sine (58) */
  COSINE, /**< consine (59) */
  TANGENT, /**< tangent (60) */
  COSECANT, /**< cosecant (61) */
  SECANT, /**< secant (62) */
  COTANGENT, /**< cotangent (63) */
  ARCSINE, /**< arc sine (64) */
  ARCCOSINE, /**< arc consine (65) */
  ARCTANGENT, /**< arc tangent (66) */
  ARCCOSECANT, /**< arc cosecant (67) */
  ARCSECANT, /**< arc secant (68) */
  ARCCOTANGENT, /**< arc cotangent (69) */
  SQRT, /**< square root (70) */
  DIVISIBLE_OP, /**< operator for the divisibility-by-k predicate; payload is an instance of the cvc5::internal::Divisible class (71) */
  CONST_RATIONAL, /**< a multiple-precision rational constant; payload is an instance of the cvc5::internal::Rational class (72) */
  CONST_INTEGER, /**< a multiple-precision integer constant; payload is an instance of the cvc5::internal::Rational class (73) */
  REAL_ALGEBRAIC_NUMBER_OP, /**< a real algebraic number constant; payload is an instance of the cvc5::internal::RealAlgebraicNumber class (74) */
  REAL_ALGEBRAIC_NUMBER, /**< a real algebraic number constant; payload is an instance of the cvc5::internal::RealAlgebraicNumber class (75) */
  LT, /**< less than, x < y (76) */
  LEQ, /**< less than or equal, x <= y (77) */
  GT, /**< greater than, x > y (78) */
  GEQ, /**< greater than or equal, x >= y (79) */
  INDEXED_ROOT_PREDICATE_OP, /**< operator for the indexed root predicate; payload is an instance of the cvc5::internal::IndexedRootPredicate class (80) */
  INDEXED_ROOT_PREDICATE, /**< indexed root predicate; first parameter is a INDEXED_ROOT_PREDICATE_OP, second is a real variable compared to zero, third is a polynomial (81) */
  IS_INTEGER, /**< term-is-integer predicate (parameter is a real-sorted term) (82) */
  TO_INTEGER, /**< convert term to integer by the floor function (parameter is a real-sorted term) (83) */
  TO_REAL, /**< cast term to real (parameter is an integer-sorted term) (84) */
  PI, /**< pi (85) */
  IAND_OP, /**< operator for integer AND; payload is an instance of the cvc5::internal::IntAnd class (86) */
  IAND, /**< integer version of AND operator; first parameter is an IAND_OP, second and third are integer terms (87) */

  /* from bv */
  BITVECTOR_TYPE, /**< bit-vector type (88) */
  CONST_BITVECTOR, /**< a fixed-width bit-vector constant; payload is an instance of the cvc5::internal::BitVector class (89) */
  BITVECTOR_FROM_BOOLS, /**< create bit-vector from vector of Booleans (90) */
  BITVECTOR_SIZE, /**< Size of bitvector (91) */
  CONST_BITVECTOR_SYMBOLIC, /**< A symbolic bitvector constant. The first argument is its value (an integer), and the second value is its size (also an integer). (92) */
  BITVECTOR_CONCAT, /**< concatenation of two or more bit-vectors (93) */
  BITVECTOR_AND, /**< bitwise and of two or more bit-vectors (94) */
  BITVECTOR_COMP, /**< equality comparison of two bit-vectors (returns one bit) (95) */
  BITVECTOR_OR, /**< bitwise or of two or more bit-vectors (96) */
  BITVECTOR_XOR, /**< bitwise xor of two or more bit-vectors (97) */
  BITVECTOR_NOT, /**< bitwise not of a bit-vector (98) */
  BITVECTOR_NAND, /**< bitwise nand of two bit-vectors (99) */
  BITVECTOR_NOR, /**< bitwise nor of two bit-vectors (100) */
  BITVECTOR_XNOR, /**< bitwise xnor of two bit-vectors (101) */
  BITVECTOR_MULT, /**< multiplication of two or more bit-vectors (102) */
  BITVECTOR_NEG, /**< unary negation of a bit-vector (103) */
  BITVECTOR_ADD, /**< addition of two or more bit-vectors (104) */
  BITVECTOR_SUB, /**< subtraction of two bit-vectors (105) */
  BITVECTOR_UDIV, /**< unsigned division of two bit-vectors, truncating towards 0 (defined to be the all-ones bit pattern, if divisor is 0) (106) */
  BITVECTOR_UREM, /**< unsigned remainder from truncating division of two bit-vectors (defined to be equal to the dividend, if divisor is 0) (107) */
  BITVECTOR_SDIV, /**< 2's complement signed division (108) */
  BITVECTOR_SMOD, /**< 2's complement signed remainder (sign follows divisor) (109) */
  BITVECTOR_SREM, /**< 2's complement signed remainder (sign follows dividend) (110) */
  BITVECTOR_ASHR, /**< bit-vector arithmetic shift right (the two bit-vector parameters must have same width) (111) */
  BITVECTOR_LSHR, /**< bit-vector logical shift right (the two bit-vector parameters must have same width) (112) */
  BITVECTOR_SHL, /**< bit-vector shift left (the two bit-vector parameters must have same width) (113) */
  BITVECTOR_ULE, /**< bit-vector unsigned less than or equal (the two bit-vector parameters must have same width) (114) */
  BITVECTOR_ULT, /**< bit-vector unsigned less than (the two bit-vector parameters must have same width) (115) */
  BITVECTOR_UGE, /**< bit-vector unsigned greater than or equal (the two bit-vector parameters must have same width) (116) */
  BITVECTOR_UGT, /**< bit-vector unsigned greater than (the two bit-vector parameters must have same width) (117) */
  BITVECTOR_SLE, /**< bit-vector signed less than or equal (the two bit-vector parameters must have same width) (118) */
  BITVECTOR_SLT, /**< bit-vector signed less than (the two bit-vector parameters must have same width) (119) */
  BITVECTOR_SGE, /**< bit-vector signed greater than or equal (the two bit-vector parameters must have same width) (120) */
  BITVECTOR_SGT, /**< bit-vector signed greater than (the two bit-vector parameters must have same width) (121) */
  BITVECTOR_ULTBV, /**< bit-vector unsigned less than but returns bv of size 1 instead of boolean (122) */
  BITVECTOR_SLTBV, /**< bit-vector signed less than but returns bv of size 1 instead of boolean (123) */
  BITVECTOR_REDAND, /**< bit-vector redand (124) */
  BITVECTOR_REDOR, /**< bit-vector redor (125) */
  BITVECTOR_NEGO, /**< bit-vector negation overflow predicate (126) */
  BITVECTOR_UADDO, /**< bit-vector unsigned addition overflow predicate (127) */
  BITVECTOR_SADDO, /**< bit-vector signed addition overflow predicate (128) */
  BITVECTOR_UMULO, /**< bit-vector unsigned multiplication overflow predicate (129) */
  BITVECTOR_SMULO, /**< bit-vector signed multiplication overflow predicate (130) */
  BITVECTOR_USUBO, /**< bit-vector unsigned subtraction overflow predicate (131) */
  BITVECTOR_SSUBO, /**< bit-vector signed subtraction overflow predicate (132) */
  BITVECTOR_SDIVO, /**< bit-vector signed division overflow predicate (133) */
  BITVECTOR_ITE, /**< same semantics as regular ITE, but condition is bv of size 1 instead of Boolean (134) */
  BITVECTOR_ACKERMANNIZE_UDIV, /**< term to be treated as a variable; used for eager bit-blasting Ackermann expansion of bvudiv (internal-only symbol) (135) */
  BITVECTOR_ACKERMANNIZE_UREM, /**< term to be treated as a variable; used for eager bit-blasting Ackermann expansion of bvurem (internal-only symbol) (136) */
  BITVECTOR_EAGER_ATOM, /**< formula to be treated as a bv atom via eager bit-blasting (internal-only symbol) (137) */
  BITVECTOR_BIT_OP, /**< operator for the bit-vector boolean bit extract; payload is an instance of the cvc5::internal::BitVectorBit class (138) */
  BITVECTOR_BIT, /**< bit-vector boolean bit extract; first parameter is a BITVECTOR_BIT_OP, second is a bit-vector term (139) */
  BITVECTOR_EXTRACT_OP, /**< operator for the bit-vector extract; payload is an instance of the cvc5::internal::BitVectorExtract class (140) */
  BITVECTOR_EXTRACT, /**< bit-vector extract; first parameter is a BITVECTOR_EXTRACT_OP, second is a bit-vector term (141) */
  BITVECTOR_REPEAT_OP, /**< operator for the bit-vector repeat; payload is an instance of the cvc5::internal::BitVectorRepeat class (142) */
  BITVECTOR_REPEAT, /**< bit-vector repeat; first parameter is a BITVECTOR_REPEAT_OP, second is a bit-vector term (143) */
  BITVECTOR_ROTATE_LEFT_OP, /**< operator for the bit-vector rotate left; payload is an instance of the cvc5::internal::BitVectorRotateLeft class (144) */
  BITVECTOR_ROTATE_LEFT, /**< bit-vector rotate left; first parameter is a BITVECTOR_ROTATE_LEFT_OP, second is a bit-vector term (145) */
  BITVECTOR_ROTATE_RIGHT_OP, /**< operator for the bit-vector rotate right; payload is an instance of the cvc5::internal::BitVectorRotateRight class (146) */
  BITVECTOR_ROTATE_RIGHT, /**< bit-vector rotate right; first parameter is a BITVECTOR_ROTATE_RIGHT_OP, second is a bit-vector term (147) */
  BITVECTOR_SIGN_EXTEND_OP, /**< operator for the bit-vector sign-extend; payload is an instance of the cvc5::internal::BitVectorSignExtend class (148) */
  BITVECTOR_SIGN_EXTEND, /**< bit-vector sign-extend; first parameter is a BITVECTOR_SIGN_EXTEND_OP, second is a bit-vector term (149) */
  BITVECTOR_ZERO_EXTEND_OP, /**< operator for the bit-vector zero-extend; payload is an instance of the cvc5::internal::BitVectorZeroExtend class (150) */
  BITVECTOR_ZERO_EXTEND, /**< bit-vector zero-extend; first parameter is a BITVECTOR_ZERO_EXTEND_OP, second is a bit-vector term (151) */

  /* from ff */
  FINITE_FIELD_TYPE, /**< finite field type (152) */
  CONST_FINITE_FIELD, /**< a finite-field constant; payload is an instance of the cvc5::internal::FiniteFieldValue class (153) */
  FINITE_FIELD_MULT, /**< multiplication of two or more field elements (154) */
  FINITE_FIELD_NEG, /**< unary negation of a field element (155) */
  FINITE_FIELD_ADD, /**< addition of two or more field elements (156) */
  FINITE_FIELD_BITSUM, /**< bitsum of field elements (157) */

  /* from fp */
  CONST_FLOATINGPOINT, /**< a floating-point literal (158) */
  CONST_ROUNDINGMODE, /**< a floating-point rounding mode (159) */
  FLOATINGPOINT_TYPE, /**< floating-point type (160) */
  FLOATINGPOINT_FP, /**< construct a floating-point literal from bit vectors (161) */
  FLOATINGPOINT_EQ, /**< floating-point equality (162) */
  FLOATINGPOINT_ABS, /**< floating-point absolute value (163) */
  FLOATINGPOINT_NEG, /**< floating-point negation (164) */
  FLOATINGPOINT_ADD, /**< floating-point addition (165) */
  FLOATINGPOINT_SUB, /**< floating-point sutraction (166) */
  FLOATINGPOINT_MULT, /**< floating-point multiply (167) */
  FLOATINGPOINT_DIV, /**< floating-point division (168) */
  FLOATINGPOINT_FMA, /**< floating-point fused multiply and add (169) */
  FLOATINGPOINT_SQRT, /**< floating-point square root (170) */
  FLOATINGPOINT_REM, /**< floating-point remainder (171) */
  FLOATINGPOINT_RTI, /**< floating-point round to integral (172) */
  FLOATINGPOINT_MIN, /**< floating-point minimum (173) */
  FLOATINGPOINT_MAX, /**< floating-point maximum (174) */
  FLOATINGPOINT_MIN_TOTAL, /**< floating-point minimum (defined for all inputs) (175) */
  FLOATINGPOINT_MAX_TOTAL, /**< floating-point maximum (defined for all inputs) (176) */
  FLOATINGPOINT_LEQ, /**< floating-point less than or equal (177) */
  FLOATINGPOINT_LT, /**< floating-point less than (178) */
  FLOATINGPOINT_GEQ, /**< floating-point greater than or equal (179) */
  FLOATINGPOINT_GT, /**< floating-point greater than (180) */
  FLOATINGPOINT_IS_NORMAL, /**< floating-point is normal (181) */
  FLOATINGPOINT_IS_SUBNORMAL, /**< floating-point is sub-normal (182) */
  FLOATINGPOINT_IS_ZERO, /**< floating-point is zero (183) */
  FLOATINGPOINT_IS_INF, /**< floating-point is infinite (184) */
  FLOATINGPOINT_IS_NAN, /**< floating-point is NaN (185) */
  FLOATINGPOINT_IS_NEG, /**< floating-point is negative (186) */
  FLOATINGPOINT_IS_POS, /**< floating-point is positive (187) */
  FLOATINGPOINT_TO_FP_FROM_IEEE_BV_OP, /**< operator for to_fp from bit vector (188) */
  FLOATINGPOINT_TO_FP_FROM_IEEE_BV, /**< convert an IEEE-754 bit vector to floating-point (189) */
  FLOATINGPOINT_TO_FP_FROM_FP_OP, /**< operator for to_fp from floating point (190) */
  FLOATINGPOINT_TO_FP_FROM_FP, /**< convert between floating-point sorts (191) */
  FLOATINGPOINT_TO_FP_FROM_REAL_OP, /**< operator for to_fp from real (192) */
  FLOATINGPOINT_TO_FP_FROM_REAL, /**< convert a real to floating-point (193) */
  FLOATINGPOINT_TO_FP_FROM_SBV_OP, /**< operator for to_fp from signed bit vector (194) */
  FLOATINGPOINT_TO_FP_FROM_SBV, /**< convert a signed bit vector to floating-point (195) */
  FLOATINGPOINT_TO_FP_FROM_UBV_OP, /**< operator for to_fp from unsigned bit vector (196) */
  FLOATINGPOINT_TO_FP_FROM_UBV, /**< convert an unsigned bit vector to floating-point (197) */
  FLOATINGPOINT_TO_UBV_OP, /**< operator for to_ubv (198) */
  FLOATINGPOINT_TO_UBV, /**< convert a floating-point value to an unsigned bit vector (199) */
  FLOATINGPOINT_TO_UBV_TOTAL_OP, /**< operator for to_ubv_total (200) */
  FLOATINGPOINT_TO_UBV_TOTAL, /**< convert a floating-point value to an unsigned bit vector (defined for all inputs) (201) */
  FLOATINGPOINT_TO_SBV_OP, /**< operator for to_sbv (202) */
  FLOATINGPOINT_TO_SBV, /**< convert a floating-point value to a signed bit vector (203) */
  FLOATINGPOINT_TO_SBV_TOTAL_OP, /**< operator for to_sbv_total (204) */
  FLOATINGPOINT_TO_SBV_TOTAL, /**< convert a floating-point value to a signed bit vector (defined for all inputs) (205) */
  FLOATINGPOINT_TO_REAL, /**< floating-point to real (206) */
  FLOATINGPOINT_TO_REAL_TOTAL, /**< floating-point to real (defined for all inputs) (207) */
  FLOATINGPOINT_COMPONENT_NAN, /**< NaN component of a word-blasted floating-point number (208) */
  FLOATINGPOINT_COMPONENT_INF, /**< Inf component of a word-blasted floating-point number (209) */
  FLOATINGPOINT_COMPONENT_ZERO, /**< Zero component of a word-blasted floating-point number (210) */
  FLOATINGPOINT_COMPONENT_SIGN, /**< Sign component of a word-blasted floating-point number (211) */
  FLOATINGPOINT_COMPONENT_EXPONENT, /**< Exponent component of a word-blasted floating-point number (212) */
  FLOATINGPOINT_COMPONENT_SIGNIFICAND, /**< Significand component of a word-blasted floating-point number (213) */
  ROUNDINGMODE_BITBLAST, /**< The bit-vector for a non-deterministic rounding mode (214) */

  /* from arrays */
  ARRAY_TYPE, /**< array type (215) */
  SELECT, /**< array select; first parameter is an array term, second is the selection index (216) */
  STORE, /**< array store; first parameter is an array term, second is the store index, third is the term to store at the index (217) */
  EQ_RANGE, /**< equality of two arrays over an index range lower/upper (218) */
  STORE_ALL, /**< array store-all; payload is an instance of the cvc5::internal::ArrayStoreAll class (this is not supported by arrays decision procedure yet, but it is used for returned array models) (219) */
  ARRAY_LAMBDA, /**< array lambda (internal-only symbol) (220) */

  /* from datatypes */
  CONSTRUCTOR_TYPE, /**< constructor (221) */
  SELECTOR_TYPE, /**< selector (222) */
  TESTER_TYPE, /**< tester (223) */
  UPDATER_TYPE, /**< datatype update (224) */
  APPLY_CONSTRUCTOR, /**< constructor application; first parameter is the constructor, remaining parameters (if any) are parameters to the constructor (225) */
  APPLY_SELECTOR, /**< selector application; parameter is a datatype term (undefined if mis-applied) (226) */
  APPLY_TESTER, /**< tester application; first parameter is a tester, second is a datatype term (227) */
  APPLY_UPDATER, /**< datatype updater application; first parameter is an update, second is a datatype term to update, third is the value to update with (228) */
  DATATYPE_TYPE, /**<  (229) */
  PARAMETRIC_DATATYPE, /**< parametric datatype (230) */
  TUPLE_TYPE, /**< tuple type (231) */
  APPLY_TYPE_ASCRIPTION, /**< type ascription, for datatype constructor applications; first parameter is an ASCRIPTION_TYPE, second is the datatype constructor application being ascribed (232) */
  ASCRIPTION_TYPE, /**< a type parameter for type ascription; payload is an instance of the cvc5::internal::AscriptionType class (233) */
  DT_SIZE, /**< datatypes size (234) */
  DT_HEIGHT_BOUND, /**< datatypes height bound (235) */
  DT_SIZE_BOUND, /**< datatypes height bound (236) */
  DT_SYGUS_BOUND, /**< datatypes sygus bound (237) */
  DT_SYGUS_EVAL, /**< datatypes sygus evaluation function (238) */
  MATCH, /**< match construct (239) */
  MATCH_CASE, /**< a match case (240) */
  MATCH_BIND_CASE, /**< a match case with bound variables (241) */
  TUPLE_PROJECT_OP, /**< operator for TUPLE_PROJECT; payload is an instance of the cvc5::internal::ProjectOp class (242) */
  TUPLE_PROJECT, /**< projects a tuple from an existing tuple using indices passed in TupleProjectOp (243) */
  CODATATYPE_BOUND_VARIABLE, /**< the kind of expressions representing bound variables in codatatype constants, which are de Bruijn indexed variables; payload is an instance of the cvc5::internal::CodatatypeBoundVariable class (used in models) (244) */
  NULLABLE_TYPE, /**< nullable type (245) */
  NULLABLE_LIFT, /**< application of lift operator; first parameter is the lift operator, remaining ones are its arguments (246) */

  /* from sep */
  SEP_NIL, /**< separation nil (247) */
  SEP_EMP, /**< separation logic empty heap constraint (248) */
  SEP_PTO, /**< points to relation (249) */
  SEP_STAR, /**< separation star (250) */
  SEP_WAND, /**< separation magic wand (251) */
  SEP_LABEL, /**< separation label (internal use only) (252) */

  /* from sets */
  SET_EMPTY, /**< the empty set constant; payload is an instance of the cvc5::internal::EmptySet class (253) */
  SET_TYPE, /**< set type, takes as parameter the type of the elements (254) */
  SET_UNION, /**< set union (255) */
  SET_INTER, /**< set intersection (256) */
  SET_MINUS, /**< set subtraction (257) */
  SET_SUBSET, /**< subset predicate; first parameter a subset of second (258) */
  SET_MEMBER, /**< set membership predicate; first parameter a member of second (259) */
  SET_SINGLETON, /**< constructs a set of a single element. First parameter is a term (260) */
  SET_INSERT, /**< set obtained by inserting elements (first N-1 parameters) into a set (the last parameter) (261) */
  SET_CARD, /**< set cardinality operator (262) */
  SET_COMPLEMENT, /**< set complement (with respect to finite universe) (263) */
  SET_UNIVERSE, /**< (finite) universe set, all set variables must be interpreted as subsets of it. (264) */
  SET_COMPREHENSION, /**< set comprehension specified by a bound variable list, a predicate, and a term. (265) */
  SET_CHOOSE, /**< return an element in the set given as a parameter (266) */
  SET_IS_EMPTY, /**< return whether the given set is empty (267) */
  SET_IS_SINGLETON, /**< return whether the given set is a singleton (268) */
  SET_MAP, /**< set map function (269) */
  SET_FILTER, /**< set filter operator (270) */
  SET_ALL, /**< set all operator (271) */
  SET_SOME, /**< set some operator (272) */
  SET_FOLD, /**< set fold operator (273) */
  RELATION_GROUP_OP, /**< operator for RELATION_GROUP; payload is an instance of the cvc5::internal::ProjectOp class (274) */
  RELATION_GROUP, /**< relation group (275) */
  RELATION_AGGREGATE_OP, /**< operator for RELATION_AGGREGATE; payload is an instance of the cvc5::internal::ProjectOp class (276) */
  RELATION_AGGREGATE, /**< relation aggregate (277) */
  RELATION_PROJECT_OP, /**< operator for RELATION_PROJECT; payload is an instance of the cvc5::internal::ProjectOp class (278) */
  RELATION_PROJECT, /**< relation projection (279) */
  RELATION_JOIN, /**< relation join (280) */
  RELATION_PRODUCT, /**< relation cartesian product (281) */
  RELATION_TRANSPOSE, /**< relation transpose (282) */
  RELATION_TCLOSURE, /**< relation transitive closure (283) */
  RELATION_JOIN_IMAGE, /**< relation join image (284) */
  RELATION_IDEN, /**< relation identity (285) */
  RELATION_TABLE_JOIN_OP, /**< operator for RELATION_TABLE_JOIN; payload is an instance of the cvc5::internal::ProjectOp class (286) */
  RELATION_TABLE_JOIN, /**< relation table join (287) */
  SET_EMPTY_OF_TYPE, /**< the empty set whose argument represents its type (288) */

  /* from bags */
  BAG_EMPTY, /**< the empty bag constant; payload is an instance of the cvc5::internal::EmptyBag class (289) */
  BAG_TYPE, /**< bag type, takes as parameter the type of the elements (290) */
  BAG_UNION_MAX, /**< union for bags (max) (291) */
  BAG_UNION_DISJOINT, /**< disjoint union for bags (sum) (292) */
  BAG_INTER_MIN, /**< bag intersection (min) (293) */
  BAG_DIFFERENCE_SUBTRACT, /**< bag difference1 (subtracts multiplicities) (294) */
  BAG_DIFFERENCE_REMOVE, /**< bag difference remove (removes shared elements) (295) */
  BAG_SUBBAG, /**< inclusion predicate for bags (less than or equal multiplicities) (296) */
  BAG_COUNT, /**< multiplicity of an element in a bag (297) */
  BAG_MEMBER, /**< bag membership predicate; is first parameter a member of second? (298) */
  BAG_SETOF, /**< eliminate duplicates in a bag (also known as the delta operator,or the squash operator) (299) */
  BAG_MAKE, /**< constructs a bag from one element along with its multiplicity (300) */
  BAG_CARD, /**< bag cardinality operator (301) */
  BAG_CHOOSE, /**< return an element in the bag given as a parameter (302) */
  BAG_MAP, /**< bag map function (303) */
  BAG_FILTER, /**< bag filter operator (304) */
  BAG_ALL, /**< bag all operator (305) */
  BAG_SOME, /**< bag some operator (306) */
  BAG_FOLD, /**< bag fold operator (307) */
  BAG_PARTITION, /**< bag partition operator (308) */
  TABLE_PRODUCT, /**< table cross product (309) */
  TABLE_PROJECT_OP, /**< operator for TABLE_PROJECT; payload is an instance of the cvc5::internal::ProjectOp class (310) */
  TABLE_PROJECT, /**< table projection (311) */
  TABLE_AGGREGATE_OP, /**< operator for TABLE_AGGREGATE; payload is an instance of the cvc5::internal::ProjectOp class (312) */
  TABLE_AGGREGATE, /**< table aggregate (313) */
  TABLE_JOIN_OP, /**< operator for TABLE_JOIN; payload is an instance of the cvc5::internal::ProjectOp class (314) */
  TABLE_JOIN, /**< table join (315) */
  TABLE_GROUP_OP, /**< operator for TABLE_GROUP; payload is an instance of the cvc5::internal::ProjectOp class (316) */
  TABLE_GROUP, /**< table group (317) */

  /* from strings */
  STRING_CONCAT, /**< string concat (N-ary) (318) */
  STRING_IN_REGEXP, /**< membership (319) */
  STRING_LENGTH, /**< string length (320) */
  STRING_SUBSTR, /**< string substr (321) */
  STRING_UPDATE, /**< string update (322) */
  STRING_CHARAT, /**< string charat (323) */
  STRING_CONTAINS, /**< string contains (324) */
  STRING_LT, /**< string less than (325) */
  STRING_LEQ, /**< string less than or equal (326) */
  STRING_INDEXOF, /**< string index of substring (327) */
  STRING_INDEXOF_RE, /**< string index of regular expression match (328) */
  STRING_REPLACE, /**< string replace (329) */
  STRING_REPLACE_ALL, /**< string replace all (330) */
  STRING_REPLACE_RE, /**< string replace regular expression match (331) */
  STRING_REPLACE_RE_ALL, /**< string replace all regular expression matches (332) */
  STRING_PREFIX, /**< string prefixof (333) */
  STRING_SUFFIX, /**< string suffixof (334) */
  STRING_IS_DIGIT, /**< string isdigit, returns true if argument is a string of length one that represents a digit (335) */
  STRING_ITOS, /**< integer to string (336) */
  STRING_STOI, /**< string to integer (total function) (337) */
  STRING_TO_CODE, /**< string to code, returns the code of the first character of the string if it has length one, -1 otherwise (338) */
  STRING_FROM_CODE, /**< string from code, returns a string containing a single character whose code point matches the argument to this function, empty string if the argument is not a valid code point (339) */
  STRING_UNIT, /**< string unit, returns a string containing a single character whose code point matches the argument to this function, arbitrary string of length one if the argument is not a valid code point (340) */
  STRING_TO_LOWER, /**< string to lowercase conversion (341) */
  STRING_TO_UPPER, /**< string to uppercase conversion (342) */
  STRING_REV, /**< string reverse (343) */
  CONST_STRING, /**< a string of characters (344) */
  SEQUENCE_TYPE, /**< sequence type, takes as parameter the type of the elements (345) */
  CONST_SEQUENCE, /**< a sequence of characters (346) */
  SEQ_UNIT, /**< a sequence of length one. First parameter is a term (347) */
  SEQ_NTH, /**< The nth element of a sequence (348) */
  STRING_TO_REGEXP, /**< convert string to regexp (349) */
  REGEXP_CONCAT, /**< regexp concat (350) */
  REGEXP_UNION, /**< regexp union (351) */
  REGEXP_INTER, /**< regexp intersection (352) */
  REGEXP_DIFF, /**< regexp difference (353) */
  REGEXP_STAR, /**< regexp * (354) */
  REGEXP_PLUS, /**< regexp + (355) */
  REGEXP_OPT, /**< regexp ? (356) */
  REGEXP_RANGE, /**< regexp range (357) */
  REGEXP_COMPLEMENT, /**< regexp complement (358) */
  REGEXP_NONE, /**< regexp empty (359) */
  REGEXP_ALL, /**< regexp all (360) */
  REGEXP_ALLCHAR, /**< regexp all characters (361) */
  REGEXP_REPEAT_OP, /**< operator for regular expression repeat; payload is an instance of the cvc5::internal::RegExpRepeat class (362) */
  REGEXP_REPEAT, /**< regular expression repeat; first parameter is a REGEXP_REPEAT_OP, second is a regular expression term (363) */
  REGEXP_LOOP_OP, /**< operator for regular expression loop; payload is an instance of the cvc5::internal::RegExpLoop class (364) */
  REGEXP_LOOP, /**< regular expression loop; first parameter is a REGEXP_LOOP_OP, second is a regular expression term (365) */
  REGEXP_RV, /**< regexp rv (internal use only) (366) */
  SEQ_EMPTY_OF_TYPE, /**< the empty sequence whose argument represents its type (367) */

  /* from quantifiers */
  FORALL, /**< universally quantified formula; first parameter is an BOUND_VAR_LIST, second is quantifier body, and an optional third parameter is an INST_PATTERN_LIST (368) */
  EXISTS, /**< existentially quantified formula; first parameter is an BOUND_VAR_LIST, second is quantifier body, and an optional third parameter is an INST_PATTERN_LIST (369) */
  INST_CONSTANT, /**< instantiation constant (370) */
  ORACLE, /**< oracle (371) */
  BOUND_VAR_LIST, /**< a list of bound variables (used to bind variables under a quantifier) (372) */
  INST_PATTERN, /**< instantiation pattern (373) */
  INST_NO_PATTERN, /**< instantiation no-pattern (374) */
  INST_ATTRIBUTE, /**< instantiation attribute (375) */
  INST_POOL, /**< instantiation pool (376) */
  INST_ADD_TO_POOL, /**< instantiation add to pool (377) */
  SKOLEM_ADD_TO_POOL, /**< skolemization add to pool (378) */
  ORACLE_FORMULA_GEN, /**< oracle interface (379) */
  INST_PATTERN_LIST, /**< a list of instantiation patterns (380) */

 LAST_KIND /**< marks the upper-bound of this enumeration */
  // clang-format on

}; /* enum Kind_t */

}  // namespace kind

// import Kind into the "cvc5" namespace but keep the individual kind
// constants under kind::
typedef cvc5::internal::kind::Kind_t Kind;

namespace kind {

/**
 * Converts an kind to a string. Note: This function is also used in
 * `safe_print()`. Changing this functions name or signature will result in
 * `safe_print()` printing "<unsupported>" instead of the proper strings for
 * the enum values.
 *
 * @param k The kind
 * @return The name of the kind
 */
const char* toString(cvc5::internal::Kind k);

/**
 * Writes a kind name to a stream.
 *
 * @param out The stream to write to
 * @param k The kind to write to the stream
 * @return The stream
 */
std::ostream& operator<<(std::ostream&, cvc5::internal::Kind);

/** Returns true if the given kind is associative. This is used by ExprManager to
 * decide whether it's safe to modify big expressions by changing the grouping of
 * the arguments. */
/* TODO: This could be generated. */
bool isAssociative(cvc5::internal::Kind k);
std::string kindToString(cvc5::internal::Kind k);

/** Return true if k is a closure kind. */
bool isClosureKind(cvc5::internal::Kind k);

struct KindHashFunction
{
  inline size_t operator()(cvc5::internal::Kind k) const
  {
    return static_cast<size_t>(k);
  }
}; /* struct KindHashFunction */

}  // namespace kind

/**
 * The enumeration for the built-in atomic types.
 */
enum TypeConstant
{
  // clang-format off
    BUILTIN_OPERATOR_TYPE, /**< the type for built-in operators*/
  SEXPR_TYPE, /**< the type of a symbolic expression*/
  BOOLEAN_TYPE, /**< Boolean type*/
  REAL_TYPE, /**< real type*/
  INTEGER_TYPE, /**< integer type*/
  ROUNDINGMODE_TYPE, /**< floating-point rounding mode*/
  STRING_TYPE, /**< String type*/
  REGEXP_TYPE, /**< RegExp type*/
  BOUND_VAR_LIST_TYPE, /**< the type of bound variable lists*/
  INST_PATTERN_TYPE, /**< instantiation pattern type*/
  INST_PATTERN_LIST_TYPE, /**< the type of instantiation pattern lists*/
 LAST_TYPE
  // clang-format on
}; /* enum TypeConstant */

/**
 * We hash the constants with their values.
 */
struct TypeConstantHashFunction
{
  inline size_t operator()(TypeConstant tc) const { return tc; }
}; /* struct TypeConstantHashFunction */

const char* toString(TypeConstant tc);
std::ostream& operator<<(std::ostream& out, TypeConstant typeConstant);

namespace theory {

cvc5::internal::theory::TheoryId kindToTheoryId(cvc5::internal::Kind k);
cvc5::internal::theory::TheoryId typeConstantToTheoryId(
    cvc5::internal::TypeConstant typeConstant);

}  // namespace theory
}  // namespace cvc5::internal

#endif /* CVC5__KIND_H */
